from pydub import AudioSegment

from inoft_vocal_framework.audio_editing.sound import Sound
from inoft_vocal_framework.audio_editing.track import Track
from inoft_vocal_framework.audio_editing.relation import Relation
from inoft_vocal_framework.dummy_object import DummyObject
from typing import List, Optional


class AudioClip:
    def __init__(self):
        self.tracks: List[Track] = list()

    def _render(self) -> AudioSegment:
        if len(self.tracks) > 1:
            id_longer_primary_track = None
            duration_longer_primary_track = 0

            for i, track in enumerate(self.tracks):
                track._audio_segment = track._render()
                if track.is_primary:
                    if track._audio_segment.duration_seconds > duration_longer_primary_track:
                        id_longer_primary_track = i
                        duration_longer_primary_track = track._audio_segment.duration_seconds

            if id_longer_primary_track is not None:
                last_overlayed_audio = self.tracks[id_longer_primary_track]._audio_segment
                self.tracks.pop(id_longer_primary_track)
            else:
                last_overlayed_audio = self.tracks[0]._audio_segment
                self.tracks.pop(0)

            for track in self.tracks:
                last_overlayed_audio = last_overlayed_audio.overlay(track._audio_segment, loop=track.loop_until_primary_tracks_finish)
            return last_overlayed_audio
        elif len(self.tracks) > 0:
            return self.tracks[0]._audio_segment
        else:
            return None

    def _export(self, filepath: str, format_type: str = "mp3"):
        audio_segment_result = self._render()
        if audio_segment_result is not None:
            audio_segment_result.export(out_f=filepath, format=format_type)

    def add_track(self, track: Track):
        if not isinstance(track, Track):
            raise Exception(f"You can only add Track objects to an AudioClip but you tried to add {track}")
        self.tracks.append(track)

    def __add__(self, track: Track):
        self.add_track(track=track)

    def track(self, name: str) -> Track:
        for track in self.tracks:
            if track.name == name:
                return track
        return DummyObject()


if __name__ == "__main__":
    audioclip = AudioClip()

    track_1 = Track(is_primary=True)

    river_track = Track(is_primary=False, loop_until_primary_tracks_finish=True)
    river_background = Sound(local_filepath="F:/Sons utiles/2009/LucasFilm Sound Effects Library/LucasFilm Sound Effects Library CD2 - Elements/track 43.mp3")
    river_background.change_volume(-6.0)
    river_track.append_sound(river_background)

    forest_track = Track(is_primary=False, loop_until_primary_tracks_finish=True)
    forest_background = Sound(local_filepath="F:/Sons utiles/2009/LucasFilm Sound Effects Library/LucasFilm Sound Effects Library CD1  - Animal Sounds/track 95.mp3")
    forest_background.change_volume(-6.0)
    forest_track.append_sound(forest_background)

    background_music_track = Track(is_primary=False, loop_until_primary_tracks_finish=False)
    background_music = Sound(local_filepath="F:/Sons utiles/Musics/Vintage (1940s) French Music/Plus Rien N'Existe - Jean Sablon - Wal-Berg.mp3")
    background_music.change_volume(-8.0)
    background_music_track.append_sound(background_music)

    sound_effects_tracks = Track(is_primary=False, loop_until_primary_tracks_finish=False)
    walking_on_dirt_1 = Sound(local_filepath="F:/Sons utiles/Sound Effects/Walks/407659__nagwense__soft-shoes-walking-on-dirt-road.wav")
    sound_effects_tracks.append_sound(walking_on_dirt_1)

    audioclip.add_track(river_track)
    audioclip.add_track(forest_track)
    audioclip.add_track(background_music_track)
    audioclip.add_track(sound_effects_tracks)
    audioclip._export("F:/Sons utiles/test1.mp3")


    """
    t2 = Track(is_primary=True, loop_until_primary_tracks_finish=True)
    s4 = Sound(local_filepath="F:/Sons utiles/2007/Magix SoundPool Collection/NuMetal_2/Synth/arabiclead3.wav")
    s4.change_volume(-40.0)
    t2.append_sound(s4)

    t1 = Track(is_primary=False)
    s3 = Sound(local_filepath="F:/Sons utiles/2009/LucasFilm Sound Effects Library/LucasFilm Sound Effects Library CD2 - Elements/Track 07 - heavy rain.mp3")
    s3.relation = Relation(parent_sound=s4, seconds_child_start_after_parent_end=-3)
    t1.append_sound(s3)

    a = AudioClip()
    a.add_track(t2)
    a.add_track(t1)
    print(a._export(local_filepath="F:/Sons utiles/Sound Effects/dummy_test2.mp3"))
    """




from .skill_builder.inoft_skill_builder import InoftSkill, InoftCondition, InoftHandler,\
    InoftRequestHandler, InoftStateHandler, InoftDefaultFallback, InoftHandlersGroup

from .platforms_handlers.simulator.simulator_core import Simulator
from .speechs.ssml_builder_core import Speech, SpeechsList
from .audio_editing.audioclip import AudioClip, Sound, Track, Relation

# todo: if the framework do not find the AWS credentials, ask with a dialog
#  if the developer want to run the app with the database disabled



from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type, raise_if_variable_not_expected_type_and_not_none


class RelationProps:
    def __init__(self):
        self._parent_sound = None
        self._seconds_child_start_after_parent_start = None
        self._seconds_child_start_after_parent_end = None
        self._seconds_child_start_after_parent_start = None
        self._seconds_child_end_after_parent_end = None

    @property
    def parent_sound(self):
        return self._parent_sound

    @parent_sound.setter
    def parent_sound(self, parent_sound) -> None:
        self._parent_sound = parent_sound

    @property
    def seconds_child_start_after_parent_start(self) -> int:
        return self._seconds_child_start_after_parent_start

    @seconds_child_start_after_parent_start.setter
    def seconds_child_start_after_parent_start(self, seconds_child_start_after_parent_start: int) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=seconds_child_start_after_parent_start, expected_type=int,
                                                         variable_name="seconds_child_start_after_parent_start")
        self._seconds_child_start_after_parent_start = seconds_child_start_after_parent_start

    @property
    def seconds_child_start_after_parent_end(self) -> int:
        return self._seconds_child_start_after_parent_end

    @seconds_child_start_after_parent_end.setter
    def seconds_child_start_after_parent_end(self, seconds_child_start_after_parent_end: int) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=seconds_child_start_after_parent_end, expected_type=int,
                                                         variable_name="seconds_child_start_after_parent_end")
        self._seconds_child_start_after_parent_end = seconds_child_start_after_parent_end

    @property
    def seconds_child_start_after_parent_end(self) -> int:
        return self._seconds_child_start_after_parent_end

    @seconds_child_start_after_parent_end.setter
    def seconds_child_start_after_parent_end(self, seconds_child_start_after_parent_end: int) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=seconds_child_start_after_parent_end, expected_type=int,
                                                         variable_name="seconds_child_start_after_parent_end")
        self._seconds_child_start_after_parent_end = seconds_child_start_after_parent_end

    @property
    def seconds_child_end_after_parent_end(self) -> int:
        return self._seconds_child_end_after_parent_end

    @seconds_child_end_after_parent_end.setter
    def seconds_child_end_after_parent_end(self, seconds_child_end_after_parent_end: int) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=seconds_child_end_after_parent_end, expected_type=int,
                                                         variable_name="seconds_child_end_after_parent_end")
        self._seconds_child_end_after_parent_end = seconds_child_end_after_parent_end


class Relation(RelationProps):
    def __init__(self, parent_sound, seconds_child_start_after_parent_start: float = None, seconds_child_start_after_parent_end: float = None,
                 seconds_child_end_after_parent_start: float = None, seconds_child_end_after_parent_end: float = None):
        super().__init__()
        self.parent_sound = parent_sound

        self.seconds_child_start_after_parent_start = seconds_child_start_after_parent_start
        self.seconds_child_start_after_parent_end = seconds_child_start_after_parent_end
        self.seconds_child_end_after_parent_start = seconds_child_end_after_parent_start
        self.seconds_child_end_after_parent_end = seconds_child_end_after_parent_end
        grouped_events = {"seconds_child_start_after_parent_start": self.seconds_child_start_after_parent_start,
                          "seconds_child_start_after_parent_end": self.seconds_child_start_after_parent_end,
                          "seconds_child_end_after_parent_start": self.seconds_child_end_after_parent_start,
                          "seconds_child_end_after_parent_end": self.seconds_child_end_after_parent_end}
        if self._count_times_value_in_list_not_none(grouped_events) > 1:
            raise Exception(f"For every relation, you can  set a value on only one of the following events {grouped_events}")

    @staticmethod
    def _count_times_value_in_list_not_none(list_or_dict) -> int:
        list_object = list_or_dict if isinstance(list_or_dict, list) else (list_or_dict.values() if isinstance(list_or_dict, dict) else list())
        count = 0
        for item in list_object:
            if item is not None:
                count += 1
        return count





import json
from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict

filepath = "/inoft_vocal_framework/platforms_handlers/simulator/request_samples/google/launch.json"
with open(filepath, "r") as file:
    event_str = json.load(file)
    event_body_dict = NestedObjectToDict.get_dict_from_json(stringed_json_dict=event_str["body"])
    print(json.dumps(event_body_dict))




import time
from concurrent.futures.thread import ThreadPoolExecutor

import boto3
import botocore
import click
from botocore.exceptions import ClientError, NoCredentialsError


class Core:
    def __init__(self):
        self.boto_session = None
        self.s3_client = None

        self.pool = ThreadPoolExecutor(2)
        # We cannot use asyncio for the initialization of the client, because currently the http
        # requests made by boto3 are blocking http request made with the default urllib library.
        # See (https://www.mathewmarcus.com/blog/asynchronous-aws-api-requests-with-asyncio.html)
        boto_session_future = self.pool.submit(self._create_boto_session)
        s3_client_future = self.pool.submit(self._create_s3_client)
        while not s3_client_future.done():
            time.sleep(0.01)

    def _create_boto_session(self):
        self.boto_session = boto3.Session()

    def _create_s3_client(self):
        self.s3_client = boto3.client("s3")

    def create_s3_bucket_if_missing(self, bucket_name: str, region_name: str):
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)
        except ClientError as e:
            click.echo(f"Trying to create a new S3 Bucket with name {click.style(text=bucket_name, fg='yellow', bold=True)}"
                       f" in region {click.style(text=region_name, fg='yellow', bold=True)}")
            available_regions_for_s3 = self.boto_session.get_available_regions(service_name="s3")
            if region_name not in available_regions_for_s3:
                raise Exception(f"The region {region_name} was not available for s3. Here is the available regions : {available_regions_for_s3}")

            self.s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={"LocationConstraint": region_name})
            click.echo(f"Completed creation of the new bucket.")

    def upload_to_s3(self, filepath: str, object_key_name: str, bucket_name: str, region_name: str) -> bool:
        # If an error happen while uploading to S3, then the upload will not be
        # successful. We use that as our way to send a success response.
        try:
            self.create_s3_bucket_if_missing(bucket_name=bucket_name, region_name=region_name)
            self.s3_client.upload_file(Filename=filepath, Bucket=bucket_name, Key=object_key_name)
            return True
        except NoCredentialsError as e:
            click.echo(f"Almost there ! You just need to configure your AWS credentials."
                       f"\nYou can follow the official documentation (you will need to install the awscli by running pip install awscli) "
                       f"then go to https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html#configuration")
        except Exception as e:
            click.echo(f"Error while getting/creating/uploading to the S3 bucket : {e}")
            return False


    def remove_from_s3(self, file_name: str, bucket_name: str):
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                return False

        try:
            self.s3_client.delete_object(Bucket=bucket_name, Key=file_name)
            return True
        except (botocore.exceptions.ParamValidationError, botocore.exceptions.ClientError):
            return False



import os
from pathlib import Path
from typing import Optional

from pydub import AudioSegment

from inoft_vocal_framework.audio_editing.relation import Relation
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type


class SoundProps:
    def __init__(self):
        pass

class Sound(SoundProps):
    def __init__(self, local_filepath: Optional[str] = None,
                 relation: Optional[Relation] = None, volume_gain: Optional[float] = 0.0,
                 source_file_s3_bucket_name: Optional[str] = None, source_file_s3_bucket_region: Optional[str] = None,
                 source_file_s3_item_path: Optional[str] = None,
                 render_file_s3_bucket_name: Optional[str] = None, render_file_s3_bucket_region: Optional[str] = None,
                 render_file_s3_item_path: Optional[str] = None):
        super().__init__()

        from inspect import stack, getmodule
        self.sound_initializer_frame = stack()[1]

        self.local_filepath = local_filepath
        if isinstance(self.local_filepath, AudioSegment):
            # If the local_filepath arg is overridden and loaded with an audio_segment object, we load him right away.
            # I prefer to override the local_filepath arg, instead of create a load_from_audio_segment function, to
            # avoid any confusions, and to make it clear this is for backend only.
            self._audio_segment = local_filepath
        else:
            if not os.path.exists(self.local_filepath):
                raise Exception(f"No file has not been found at {self.local_filepath}")
            self._audio_segment = AudioSegment.from_file(self.local_filepath, Path(self.local_filepath).suffix.replace(".", ""))

        self.change_volume(volume_gain=volume_gain)

        self._in_use = False
        self._relation = relation
        self._append_sound_after_last_one = True

        self._seconds_start_time = None
        self._seconds_end_time = None

    @property
    def seconds_start_time(self) -> float:
        if self._seconds_start_time is None and self._seconds_end_time is not None:
            self._seconds_start_time = self._seconds_end_time - self.duration_seconds
        return self._seconds_start_time

    @seconds_start_time.setter
    def seconds_start_time(self, seconds: float) -> None:
        raise_if_variable_not_expected_type(value=seconds, expected_type=float, variable_name="seconds_start_time")
        self._seconds_start_time = seconds

    @property
    def seconds_end_time(self) -> float:
        if self._seconds_end_time is None and self._seconds_start_time is not None:
            self._seconds_end_time = self._seconds_start_time + self.duration_seconds
        return self._seconds_end_time

    @seconds_end_time.setter
    def seconds_end_time(self, seconds: float) -> None:
        raise_if_variable_not_expected_type(value=seconds, expected_type=float, variable_name="seconds_end_time")
        self._seconds_end_time = seconds

    @property
    def relation(self) -> Relation:
        return self._relation

    @relation.setter
    def relation(self, relation: Relation) -> None:
        raise_if_variable_not_expected_type(value=relation, expected_type=Relation, variable_name="relation")
        self._relation = relation

    @property
    def append_sound_after_last_one(self) -> bool:
        return self._append_sound_after_last_one

    @append_sound_after_last_one.setter
    def append_sound_after_last_one(self, append_sound_after_last_one: bool) -> None:
        raise_if_variable_not_expected_type(value=append_sound_after_last_one, expected_type=bool, variable_name="append_sound_after_last_one")
        self._append_sound_after_last_one = append_sound_after_last_one

    def copy(self):
        from copy import deepcopy
        copy = deepcopy(self)
        copy._in_use = False
        return copy

    def change_volume(self, volume_gain: float = 0.0) -> None:
        if volume_gain != 0.0:
            self._audio_segment = self._audio_segment.apply_gain(volume_change=volume_gain)

    def __len__(self):
        return self._audio_segment.__len__()

    def __eq__(self, other):
        return self._audio_segment.__eq__(other=other)

    def __hash__(self):
        return self._audio_segment.__hash__()

    def __ne__(self, other):
        return self._audio_segment.__ne__(other=other)

    def __iter__(self):
        return self._audio_segment.__iter__()

    def __getitem__(self, millisecond):
        return Sound(self._audio_segment.__getitem__(millisecond=millisecond))

    def get_sample_slice(self, start_sample=None, end_sample=None):
        return self._audio_segment.get_sample_slice(start_sample=start_sample, end_sample=end_sample)

    def __add__(self, arg):
        return self._audio_segment.__add__(arg=arg)

    def __radd__(self, rarg):
        return self._audio_segment.__radd__(rarg=rarg)

    def __sub__(self, arg):
        return self._audio_segment.__sub__(arg=arg)

    def __mul__(self, arg):
        return self._audio_segment.__mul__(arg=arg)

    @property
    def duration_seconds(self):
        return self._audio_segment.duration_seconds

    def get_dc_offset(self, channel=1):
        self._audio_segment = self._audio_segment.get_dc_offset(channel=channel)
        return self

    def remove_dc_offset(self, channel=None, offset=None):
        self._audio_segment = self._audio_segment.remove_dc_offset(channel=channel, offset=offset)
        return self

    def overlay(self, seg, seconds_position=0.0, loop=False, times=None, gain_during_overlay=None):
        self._audio_segment = self._audio_segment.overlay(seg=seg, position=seconds_position * 1000, loop=loop, times=times, gain_during_overlay=gain_during_overlay)
        return self

    def append(self, seg, crossfade=100):
        self._audio_segment = self._audio_segment.append(seg=seg, crossfade=crossfade)
        return self

    def fade(self, to_gain=0, from_gain=0, start=None, end=None, duration=None):
        self._audio_segment = self._audio_segment.fade(to_gain=to_gain, from_gain=from_gain, start=start, end=end, duration=duration)
        return self

    def fade_out(self, duration):
        self._audio_segment = self._audio_segment.fade_out(duration=duration)
        return self

    def fade_in(self, duration):
        self._audio_segment = self._audio_segment.fade_in(duration=duration)
        return self

    def reverse(self):
        self._audio_segment = self._audio_segment.reverse()
        return self



import os

from inoft_vocal_framework.audio_editing.aws_clients.core import Core

a = 9

from inoft_vocal_framework.audio_editing.sound import Sound


class Interface(Core):
    def __init__(self):
        super().__init__()

    def upload_sound(self, sound: Sound):
        bucket_name = "lebucketavecbenoit"
        region_name = "eu-west-3"
        object_key_name = "test_auto_upload.mp3"

        if os.path.exists(sound.local_filepath):
            self.upload_to_s3(filepath=sound.local_filepath, object_key_name=object_key_name, bucket_name=bucket_name, region_name=region_name)

            if os.path.exists(sound.sound_initializer_frame.filename):
                with open(sound.sound_initializer_frame.filename, 'r') as file:
                    lines = file.read().split("\n")

                last_line_initialization_of_sound_object = lines[sound.sound_initializer_frame.lineno - 1]
                last_line_initialization_of_sound_object.replace(" ", "")
                if last_line_initialization_of_sound_object[-1] == ")":
                    new_last_line = last_line_initialization_of_sound_object.replace(
                        ")", f',\nsource_file_s3_bucket_name=\"{bucket_name}",\n'
                             f'source_file_s3_bucket_region="{region_name}",\n'
                             f'source_file_s3_item_path="{object_key_name}")')
                    lines[sound.sound_initializer_frame.lineno - 1] = new_last_line

                with open(sound.sound_initializer_frame.filename, "w") as file:
                    file.write("\n".join(lines))


if __name__ == "__main__":
    s = Sound(local_filepath="F:/Sons utiles/2007/Magix SoundPool Collection/NuMetal_2/Synth/arabiclead3.wav",
source_file_s3_bucket_name="lebucketavecbenoit",
source_file_s3_bucket_region="eu-west-3",
source_file_s3_item_path="test_auto_upload.mp3")
    Interface().upload_sound(s)


"""
for i in range(10):
    with open(__file__, 'r') as f:
        lines = f.read().split('\n')
        val = int(lines[20].split(' = ')[-1])
        new_line = 'a = {}'.format(i)
        new_file = '\n'.join([new_line] + lines[1:])

    with open(__file__, 'w') as f:
        f.write('\n'.join([new_line] + lines[1:]))

    print("hey")
"""


from inoft_vocal_framework.platforms_handlers import handler_input

def basic_messages():
    handler = handler_input.HandlerInput()

    handler._force_load_alexa()
    handler.say("One")
    handler.say("Two")
    assert handler.alexa.response.outputSpeech.text == "One\nTwo"

    handler._force_load_dialogflow()
    handler.say("One")
    handler.say("Two")
    assert handler.google.response.payload.google.richResponse.items[0].textToSpeech == "One\nTwo"

basic_messages()